import logging
import datetime
import inspect
from os.path import basename


__logger = None


class TrLogging(object):
    def __init__(self, log_file: str, log_level: int) -> None:
        self._logger = logging.getLogger()
        self._raw_formatter = logging.Formatter('%(message)s')
        self._default_formatter = logging.Formatter('%(levelname).1s %(message)s')
        self._file_handler = logging.FileHandler(log_file, encoding='utf-8')
        self._file_handler.setFormatter(self._default_formatter)
        self._file_handler.setLevel(log_level)
        self._logger.addHandler(self._file_handler)
        self._logger.setLevel(log_level)
        self.raw_format = False

    def set_raw_format(self) -> None:
        self.raw_format = True
        self._file_handler.setFormatter(self._raw_formatter)

    def set_default_format(self) -> None:
        self.raw_format = False
        self._file_handler.setFormatter(self._default_formatter)

    def is_enabled_for(self, level: int) -> bool:
        return self._logger.isEnabledFor(level)

    @staticmethod
    def _frame_str() -> str:
        frame = inspect.stack()[4]
        return f"[{basename(frame.filename)}:{frame.lineno}] {frame.function}".ljust(38, '.')

    def log(self, verbosity: int, msg_args) -> None:
        if not self._logger.isEnabledFor(verbosity):
            return
        if isinstance(msg_args, tuple) and len(msg_args) == 1:
            msg_args = str(msg_args[0])
        if isinstance(msg_args, str):
            if self.raw_format:
                self._logger.log(verbosity, msg_args)
            else:
                self._logger.log(verbosity, f"{self._frame_str()}: {msg_args}")
            return
        assert type(msg_args) is tuple
        if self.raw_format:
            self._logger.log(verbosity, msg_args[0].format(*msg_args[1:]))
        else:
            self._logger.log(
                verbosity, ("{}: " + msg_args[0]).format(self._frame_str(), *msg_args[1:]))


def start_raw_logging() -> None:
    global __logger
    if __logger is None:
        return
    __logger.set_raw_format()


def stop_raw_logging() -> None:
    global __logger
    if __logger is None:
        return
    __logger.set_default_format()


def raw_msg(msg) -> None:
    global __logger
    if __logger is None:
        return
    __logger.set_raw_format()
    __logger.log(logging.INFO, msg)
    __logger.set_default_format()


def blank(count=1) -> None:
    global __logger
    if __logger is None:
        return
    __logger.set_raw_format()
    for _ in range(0, count):
        __logger.log(logging.INFO, "")
    __logger.set_default_format()


def _process_msg_arr(verbosity: int, msg_args: tuple) -> None:
    global __logger
    if __logger is None:
        return
    __logger.log(verbosity, msg_args)


def init_logging(log_file: str, logfile_verbosity: int) -> None:
    global __logger
    assert __logger is None
    if not log_file:
        return
    __logger = TrLogging(log_file, logging.INFO if logfile_verbosity == 0 else logging.DEBUG)
    blank(2)
    __logger.set_raw_format()
    info("======================================================================")
    info("taskrunner: {}".format(datetime.datetime.now().strftime("%I:%M:%S on %B %d, %Y")))
    info("======================================================================")
    __logger.set_default_format()


def logging_enabled_for(level: int) -> bool:
    global __logger
    if not __logger:
        return False
    return __logger.is_enabled_for(level)


def verbose(*msg_args) -> None:
    _process_msg_arr(logging.DEBUG, msg_args)


def info(*msg_args) -> None:
    _process_msg_arr(logging.INFO, msg_args)


def warn(*msg_args) -> None:
    _process_msg_arr(logging.WARN, msg_args)


def error(*msg_args) -> None:
    _process_msg_arr(logging.ERROR, msg_args)


def error_and_print(msg) -> None:
    print(msg)
    _process_msg_arr(logging.ERROR, msg)


def warn_and_print(msg) -> None:
    print(msg)
    _process_msg_arr(logging.WARN, msg)



from tr.config import *
from tr.schemas import TaskKeys, validate_task_schema
import shlex
import subprocess
import signal


class Task(object):
    def _check_empty_setting(self, s, title) -> None:
        if len(s) > 0:
            return
        raise TaskException(f"Expanded {title} for task '{self.name}' is empty")

    def __init__(self, name: str, config: Config) -> None:
        super().__init__()
        info("Initializing task '{}'", name)
        self.name = name
        task_descriptor = config.get_task_desc(name, True)
        validate_task_schema(name, task_descriptor)

        self.short_desc = task_descriptor.get(TaskKeys.ShortDesc, None)
        self.long_desc = task_descriptor.get(TaskKeys.LongDesc, None)
        self.hidden = task_descriptor.get(TaskKeys.Hidden, False)

        self.stop_on_error = task_descriptor.get(TaskKeys.StopOnError, True)
        self.commands = task_descriptor.get(TaskKeys.Commands, [])
        self.cwd = task_descriptor.get(TaskKeys.Cwd, None)
        self.shell = task_descriptor.get(TaskKeys.Shell, False)
        self.shell_path = task_descriptor.get(
            TaskKeys.ShellPath, config.default_shell_path())
        self.env_inherit = task_descriptor.get(TaskKeys.InheritOsEnv, True)
        self.env = task_descriptor.get(TaskKeys.Env, {})
        self.abstract = task_descriptor.get(TaskKeys.Abstract, False)

        self.c_image = task_descriptor.get(TaskKeys.CImage, None)
        self.c_volumes = task_descriptor.get(TaskKeys.CVolumes, [])
        self.c_interactive = task_descriptor.get(TaskKeys.CInteractive, False)
        self.c_tty = task_descriptor.get(TaskKeys.CTty, False)
        self.c_flags = task_descriptor.get(TaskKeys.CFlags, "")
        self.c_exec = task_descriptor.get(TaskKeys.CExec, False)
        self.c_rm = task_descriptor.get(TaskKeys.CRemove, True)
        self.c_tool = task_descriptor.get(TaskKeys.CTool, config.default_container_tool())
        self.c_shell = task_descriptor.get(TaskKeys.CShell, False)
        self.c_shell_path = task_descriptor.get(TaskKeys.CShellPath,
                                                config.default_container_shell_path())
        self.c_cwd = task_descriptor.get(TaskKeys.CCwd, None)
        self.c_env = task_descriptor.get(TaskKeys.CEnv, {})
        self.c_sudo = task_descriptor.get(TaskKeys.CSudo, False)

        self.vars_map = task_descriptor.get(TaskKeys.Variables, {})
        self.expander = None

    def expand(self) -> None:
        if self.expander is not None:
            warn("Task '{}' is already expanded", self.name)
            return
        self.expander = StringVarExpander(self.vars_map)
        info(f"Expanding task '{self.name}'")
        self.expanded = True
        self.env = {self.expander(k): self.expander(v) for k, v in self.env.items()}
        if self.cwd:
            self.cwd = self.expander(self.cwd)
        self.commands = [self.expander(c) for c in self.commands]

        if self.c_cwd:
            self.c_cwd = self.expander(self.c_cwd)
        if self.c_image:
            self.c_image = self.expander(self.c_image)
        self.c_env = {self.expander(k): self.expander(v) for k, v in self.c_env.items()}
        self.c_volumes = [self.expander(v) for v in self.c_volumes]

    def _simple_cmd_arr(self, cmd) -> list:
        info("Preparing simple command")
        if self.shell:
            return [cmd]
        try:
            return shlex.split(cmd)
        except ValueError as e:
            raise TaskException(f"Illegal command '{cmd}' for task '{self.name}' - {e}")

    def _container_cmd_arr(self, cmd) -> typing.List[str]:
        info("Preparing container command")
        cmd_array = ["sudo", self.c_tool] if self.c_sudo else [self.c_tool]
        cmd_array.append("exec" if self.c_exec else "run")
        if self.c_cwd:
            cmd_array += ["-w", self.c_cwd]
        if self.c_interactive:
            cmd_array.append("-i")
        if self.c_tty:
            cmd_array.append("-t")
        if not self.c_exec:
            if self.c_rm:
                cmd_array.append("--rm")
            for v in self.c_volumes:
                cmd_array += ["-v", v]

        for k, v in self.c_env.items():
            cmd_array += ["-e", f"{k}={v}"]
            info("Setting container environment variable will '{}={}'", k, v)

        cmd_array += self.c_flags.split()

        cmd_array.append(self.c_image)
        if self.c_shell and cmd is not None:
            cmd_array += [self.c_shell_path, "-c"]
        if cmd:
            cmd_array += [cmd]
        info("Command is {}", cmd_array)
        return cmd_array

    def _run_cmd(self, cmd: list, cmd_str: str) -> int:
        info("Running command (joined):")
        raw_msg(cmd_str)
        p = None
        if self.env_inherit:
            env = os.environ
            env.update(self.env)
        else:
            env = self.env
        try:
            p = subprocess.Popen(cmd, shell=self.shell, executable=self.shell_path, env=env,
                                 cwd=self.cwd)
            return p.wait()

        except (OSError, FileNotFoundError) as e:
            raise TaskException(f"Error occurred running command '{cmd_str}' - {e}")
        except KeyboardInterrupt:
            if p:
                p.send_signal(signal.SIGINT)
                p.wait()
            raise TaskException("User interrupt")

    def run(self) -> int:
        if not self.expanded:
            raise TaskException("Task must be expanded before run")  # Should never happen
        if self.abstract:
            raise TaskException("Can't run abstract tasks")
        if logging_enabled_for(logging.INFO):
            if self.cwd:
                info("Working directory will be set to '{}'", self.cwd)
            else:
                info("No working directory will be set")
            if self.env:
                info("Command Environment variables:")
                for k, v in self.env.items():
                    raw_msg(f"{k}={v}")

        if len(self.commands) == 0:
            if self.c_image:
                info("Running container's default command")
                return self._run_cmd(self._container_cmd_arr(None), "<CONTAINER_DEFAULT>")
            print(f"No commands defined for task '{self.name}'. Nothing to do.")
            return 0

        rc = 0
        for cmd in self.commands:
            info("Command is '{}'", cmd)
            cmd_arr = self._container_cmd_arr(cmd) if self.c_image else self._simple_cmd_arr(cmd)
            cmd_rc = self._run_cmd(cmd_arr, cmd)
            if cmd_rc == 0:
                continue
            info("Command had failed cmd_rc={}", cmd_rc)
            if self.stop_on_error:
                info("Stopping of first error")
                return cmd_rc
            if rc == 0:
                rc = cmd_rc
        return rc



from tr.schemas import *
from tr.common import *
import os
import pathlib
import json
from typing import Optional, Set
from argparse import Namespace as Args


_CONF_FILE_NAME = "tasks.json"
_DFLT_CONF_FILE_NAME = str(pathlib.Path.home()) + "/.config/" + _CONF_FILE_NAME


class Config(object):
    @staticmethod
    def _read_tasks_file(file_path: str) -> dict:
        try:
            data: dict = json.load(open(file_path, 'r'))
        except (IOError, TypeError, ValueError, TaskException) as e:
            raise TaskException(f"Error parsing {file_path} - {e}")
        return data


    def _read_configuration(self, file_path: str, read_files: set = set()) -> dict:
        conf = {}
        tasks = {}
        variables = {}
        info("Reading configuration file {}", file_path)
        orig_conf = Config._read_tasks_file(file_path)
        includes = orig_conf.get(GlobalKeys.Include, [])

        # Add the default configuration file to includes list but only for the original
        # configuration file, and the behavior isn't turned off (again, relevant ONLY to
        # original file)
        if len(read_files) == 0 and \
                file_path != _DFLT_CONF_FILE_NAME and \
                os.path.isfile(_DFLT_CONF_FILE_NAME) and \
                orig_conf.get(GlobalKeys.UseDfltInclude, True):
            includes.insert(0, _DFLT_CONF_FILE_NAME)

        info("Configuration file includes: {}", includes)
        read_files.add(file_path)
        expander = StringVarExpander()
        for f in includes:
            f = expander(f)
            if f in read_files:
                raise TaskException(f"Include loop detected - '{f}'")
            included_conf = self._read_configuration(f, read_files)
            tasks.update(included_conf[GlobalKeys.Tasks])
            variables.update(included_conf[GlobalKeys.Variables])
            conf.update(included_conf)
        read_files.remove(file_path)
        if GlobalKeys.Include in conf:
            conf.pop(GlobalKeys.Include)

        conf.update(orig_conf)
        tasks.update(orig_conf.get(GlobalKeys.Tasks, {}))
        for supressed_task in conf.get(GlobalKeys.Suppress, []):
            if supressed_task in tasks:
                info("Removing suppressed task {}", supressed_task)
                tasks.pop(supressed_task)
        conf[GlobalKeys.Tasks] = tasks
        variables.update(orig_conf.get(GlobalKeys.Variables, {}))
        conf[GlobalKeys.Variables] = variables
        return conf

    @staticmethod
    def _get_conf_file_path(cli_conf: Optional[str]) -> Optional[str]:
        if cli_conf:
            return cli_conf

        directory = os.getcwd()
        while True:
            conf_path = directory + "/." + _CONF_FILE_NAME
            if os.path.isfile(conf_path):
                return conf_path
            if directory == "/":
                break
            directory = os.path.dirname(directory)

        if os.path.isfile(_DFLT_CONF_FILE_NAME):
            return _DFLT_CONF_FILE_NAME
        return None

    def __init__(self, args: Optional[Args]) -> None:
        self.args: Args = args  # type: ignore
        conf_path = Config._get_conf_file_path(args.conf if args else None)

        #  Populate some variables early so they are available in
        const_vars = {}
        cwd = os.getcwd()
        const_vars[AutoVarsKeys.CWD] = cwd
        if conf_path:
            conf_dir_name = os.path.dirname(conf_path)
            if conf_dir_name:
                const_vars[AutoVarsKeys.TASK_ROOT] = os.path.dirname(conf_path)
        const_vars.setdefault(AutoVarsKeys.TASK_ROOT, cwd)
        if args and args.__contains__(AutoVarsKeys.TASK_CLI_ARGS):
            const_vars[AutoVarsKeys.TASK_CLI_ARGS] = " ".join(
                args.__getattribute__(AutoVarsKeys.TASK_CLI_ARGS))
        set_const_vars_map(const_vars)

        self.conf = {}
        if conf_path:
            self.conf = self._read_configuration(conf_path)
            validate_config_file_schema(self.conf)

        self.tasks = self.conf.get(GlobalKeys.Tasks, {})
        set_global_vars_map(self.conf.get(GlobalKeys.Variables, {}))

        if logging_enabled_for(logging.DEBUG):
            dump_defualt_vars()

    def default_task_name(self) -> Optional[str]:
        return self.setting(GlobalKeys.DfltTask)

    def default_container_tool(self) -> str:
        return self.setting(GlobalKeys.DfltContainerTool, default="/usr/bin/docker")

    def default_container_shell_path(self) -> str:
        return self.setting(GlobalKeys.DfltContainerShellPath, default="/usr/bin/sh")

    def default_shell_path(self) -> Optional[str]:
        return self.setting(GlobalKeys.DfltShellPath, None)

    def visible_tasks(self) -> typing.List[str]:
        tasks = set()
        for name, task in self.tasks.items():
            if not task.get(TaskKeys.Hidden, False):
                tasks.add(name)
        return list(tasks)

    #  Return anything. Types is forced by schema validations.
    def setting(self, path: str, default=None) -> typing.Any:
        return dict_value(self.conf, path, default=default)

    def _raw_task_obj(self, name: str) -> dict:
        try:
            return self.tasks[name]
        except KeyError:
            raise TaskException(f"No such task '{name}'")

    @staticmethod
    def _consolidate_commands(task_desc: dict) -> None:
            task_desc[TaskKeys.Commands] = \
               task_desc.get(TaskKeys.PreCommands, []) + \
               task_desc.get(TaskKeys.Commands, []) + \
               task_desc.get(TaskKeys.PostCommands, [])
            task_desc.pop(TaskKeys.PreCommands, None)
            task_desc.pop(TaskKeys.PostCommands, None)

    @staticmethod
    def update_task_desc(tgt_desc: dict, derived_desc:dict, merge: typing.Set[str]) -> None:
        for k, v in derived_desc.items():
            value_type = type(v)
            if k in merge and not (value_type is dict or value_type is list):
                raise TaskException(f"Unmergeable value '{k}' {type(v)}")

            if k in merge and k in tgt_desc:
                if value_type is dict:
                    if type(tgt_desc[k]) is not dict:
                        raise TaskException(f"Expecetd dictionary for key '{k}'")
                    tgt_desc[k].update(v)
                    continue
                if value_type is list:
                    if type(tgt_desc[k]) is not list:
                        raise TaskException(f"Expecetd list for key '{k}'")
                    tgt_desc[k] += v
                    continue
            tgt_desc[k] = v


    def _task_desc(self, name: str, included_list: set) -> dict:
        if name in included_list:
            raise TaskException(f"Ineritance loop detected for task '{name}'")

        included_list.add(name)
        derived_desc = self._raw_task_obj(name)

        ret_task_name = derived_desc.get(TaskKeys.Base, None)
        if ret_task_name is None or not ret_task_name.strip():
            Config._consolidate_commands(derived_desc)
            return derived_desc

        merge = set()
        if derived_desc.get(TaskKeys.InheritVariables, True):
            merge.add(TaskKeys.Variables)
        if derived_desc.get(TaskKeys.InheritEnv, True):
            merge.add(TaskKeys.Env)
        if derived_desc.get(TaskKeys.CInheritEnv, True):
            merge.add(TaskKeys.CEnv)
        if derived_desc.get(TaskKeys.CInheritVolumes, True):
            merge.add(TaskKeys.CVolumes)

        # We about to modify the included object, so deep copy it
        ret_desc = self._task_desc(ret_task_name, included_list=included_list).copy()
        for k in [TaskKeys.Hidden, TaskKeys.Abstract, TaskKeys.InheritEnv, TaskKeys.InheritVariables,
                  TaskKeys.CInheritEnv, TaskKeys.CInheritVolumes]:
            ret_desc.pop(k, None)
        Config.update_task_desc(ret_desc, derived_desc, merge)

        Config._consolidate_commands(ret_desc)
        return ret_desc

    def get_task_desc(self, name: str, includes: bool) -> typing.Dict[str, typing.Any]:
        verbose("Task '{}' requested, with_inclusions={}", name, includes)
        if includes:
            return self._task_desc(name, set())
        else:
            return self._raw_task_obj(name)
